import abc
import collections
import json
import re

from beancount.core import data
from beancount.ingest import importer
import blessed
from pytest import skip


class QueueSet:
    # A queue that bumps duplicate items to the front instead of prepending them
    def __init__(self, size):
        self.size = size
        self.queue = []

    def push(self, item):
        self.queue = [item] + list(filter(lambda i: i != item, self.queue))
        self.queue = self.queue[: self.size]

    def __iter__(self):
        return iter(self.queue)


class Event(abc.ABC):
    @abc.abstractmethod
    def get_field(self, field: str) -> str:
        pass

    @abc.abstractmethod
    def get_id(self) -> str:
        pass

    @abc.abstractmethod
    def display(self) -> str:
        pass

    @abc.abstractmethod
    def get_transaction(
        self, filename: str, index: int, recipient_account: str
    ) -> data.Transaction:
        pass


class PromptImporter(importer.ImporterProtocol, abc.ABC):
    def __init__(self, json_filename: str, skip_character: str = "x", regex_field=None):
        self.json_filename = json_filename
        self.skip_characteracter = skip_character
        self.regex_field = regex_field

    @abc.abstractmethod
    def get_events(self, f) -> list[Event]:
        pass

    def prompt(self):
        return input(">> ")

    def extract(self, f):
        def regex_matches(event: Event, field: str, regex: str) -> bool:
            r = re.compile(regex)
            return re.fullmatch(r, event.get_field(field))

        top_level_key = self.name()

        try:
            with open(self.json_filename, "r") as infile:
                mapping_data = json.load(infile)
        except FileNotFoundError:
            print("Mapping file does not exist, creating...")
            mapping_data = {}

        if top_level_key not in mapping_data:
            mapping_data[top_level_key] = {"id": [], "regex": []}

        """
        mapping_data structure:
        {
            <top_level_key>: {
                regex: [
                    {
                        field: ...,
                        regex: ...,
                        recipient ...,
                    }
                ],
                id: [
                    {
                        event_id: ...,
                        recipient: ...,
                    }
                ]
            }
        }
        """

        id_mappings = mapping_data[top_level_key]["id"]
        regex_mappings = mapping_data[top_level_key]["regex"]

        known_recipients = collections.Counter(
            [m["recipient"] for m in regex_mappings]
            + [m["recipient"] for m in id_mappings]
        )

        top_known_recipients = {}
        for index, (kr, _) in enumerate(known_recipients.most_common(3)):
            top_known_recipients[str(index + 1)] = kr

        num_top_known_recipients = len(top_known_recipients)

        recent_recipients = QueueSet(3)
        txns = []
        print_txns = True
        term = blessed.Terminal()
        for index, event in enumerate(self.get_events(f)):
            recipient_account = None

            # Check if one of the IDs matches
            for id_mapping in id_mappings:
                if event.get_id() == id_mapping["event_id"]:
                    recipient_account = id_mapping["recipient"]
                    break

            # If no ID match occurred, check if any of the regexes matches
            if recipient_account is None:
                for regex_mapping in regex_mappings:
                    if regex_matches(
                        event, regex_mapping["field"], regex_mapping["regex"]
                    ):
                        recipient_account = regex_mapping["recipient"]

            if recipient_account is None:
                print_txns = False
                self.skip_character = "x"

                print(term.home + term.clear)
                print(event.display())
                print(
                    f"What should the recipient account be? ('{self.skip_character}' to not extract a transaction)"
                )

                for rr_index, rr in enumerate(recent_recipients):
                    top_known_recipients[
                        str(num_top_known_recipients + rr_index + 1)
                    ] = rr

                known_recipients_message = ""
                for label, kr in top_known_recipients.items():
                    known_recipients_message += f"{label}. {kr}   "
                print(known_recipients_message)

                recipient_account = self.prompt().strip()

                if recipient_account in top_known_recipients:
                    recipient_account = top_known_recipients[recipient_account]

                if recipient_account == self.skip_character:
                    recipient_account = "skip"
                else:
                    recent_recipients.push(recipient_account)

                print(
                    f"What regex should identify this account (or skip) in the future? ('{self.skip_character}' to not identify this accoung with a regex)"
                )

                while True:

                    identify_regex = self.prompt().strip()

                    if identify_regex == self.skip_character:
                        id_mappings.append(
                            {"event_id": event.get_id(), "recipient": recipient_account}
                        )
                        break
                    else:
                        if self.regex_field is not None:
                            target_field = self.regex_field
                        else:
                            print(f"What field should the regex act upon?")
                            target_field = self.prompt().strip()

                        if regex_matches(event, target_field, identify_regex):
                            regex_mappings.append(
                                {
                                    "field": target_field,
                                    "regex": identify_regex,
                                    "recipient": recipient_account,
                                }
                            )
                            break
                        else:
                            print(
                                "Regex did not match this event, keep it anyway? (y/n)"
                            )
                            yes_or_no = self.prompt()
                            if yes_or_no in ["y", "Y", "yes", "Yes", "YES"]:
                                regex_mappings.append(
                                    {
                                        "field": target_field,
                                        "regex": identify_regex,
                                        "recipient": recipient_account,
                                    }
                                )
                                break

            if print_txns and recipient_account != "skip":
                txns.append(event.get_transaction(f.name, index, recipient_account))

        with open(self.json_filename, "w+") as outfile:
            json.dump(mapping_data, outfile, indent=4)

        if print_txns:
            return txns
        else:
            return []



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['prompt_importer']

package_data = \
{'': ['*']}

install_requires = \
['beancount>=2.3.5,<3.0.0', 'blessed>=1.19.1,<2.0.0']

setup_kwargs = {
    'name': 'prompt-importer',
    'version': '0.2.1',
    'description': 'A base class for building prompt-based beancount importers',
    'long_description': '# Prompt Importer\n\nThis package provides a `PromptImporter` base class to build other beancount importers with.\nThe functionality `PromptImporter` provides is to prompt the user to input recipient accounts\nfor each transaction while keeping a list of common accounts for easy access.\nFor example:\n\n```\n02/05/2022: Online payment from CHK 1234, 100.00\nWhat should the recipient account be? (\'x\' to not extract a transaction)\n1. Expenses:Groceries   2. Expenses:Restaurants   3. Assets:Checking\n>>\n```\n\nIt also provides functionality to store regexes with which to automatically identify recipient accounts in the future.\n\n## Usage\n\n### Events\n\nTo create class that derives from `PromptImporter` you must first define a subclass of `prompt_importer.Event`.\nEvents represent single transactions from whatever type of file you are importing.\nFor example, if you are importing a `csv` file, you may have one event per row.\nThe subclass should implement the following methods:\n\n#### get_field(self, field: str) -> str\n\nEach event may have different fields associated with it.\nThis method should return the value of a field given an associated field name.\n\n#### get_id(self) -> str\n\nThis should return the **globally unique** id associated with an event.\n\n#### display(self) -> str\n\nThis is how the event will be displayed to the user before the prompt.\n\n#### get_transaction(self, filename: str, index: int, recipient_account: str) -> data.Transaction\n\nThis should return a transaction associated with an event.\nTo help build the transaction, it takes the file the event was sourced from, its index within the file, and the account that should be the "recipient" of the transaction.\n\nNote that the `data.Transaction` type refers to the `data` from `beancount.core`.\n\n### Importers\n\nOnce you have defined an event you can create a subclass of `PromptImporter`.\nTo do so, you must implement the typical methods associated with the beancount`importer.ImporterProtocol` class.\n**Important:** the value the method `name(self)` returns should _not_ contain characters not allowed in SQLite table names, such as periods.\n\nThe importer should also implement the following method:\n\n#### get_events(self, f) -> list[Event]\n\nGiven a beancount file this should return a list of events for the\nimporter to process.\n\n## Example\n\nThe following is an example of an event importer for Bank of America credit card reports. The implementation of the typical beancount importer methods (`identify`, `file`, etc.) are omitted as they are not the focus.\n\n```python\nimport csv\n\nfrom prompt_importer.importer import PromptImporter, Event\n\nfrom beancount.core import amount, data, flags\nfrom beancount.core.number import D\nfrom dateutil.parser import parse\n\nclass BofaCCEvent(Event):\n    def __init__(self, row):\n        self.data = {\n            "Posted Date": row["Posted Date"],\n            "Amount": row["Amount"],\n            "Payee": row["Payee"],\n            "Reference Number": row["Reference Number"],\n        }\n\n    def get_field(self, field: str) -> str:\n        return self.data[field]\n\n    def get_id(self) -> str:\n        return f"{self.data[\'Reference Number\']}"\n\n    def display(self) -> str:\n        return (\n            f"{self.data[\'Posted Date\']}: {self.data[\'Payee\']}, {self.data[\'Amount\']}"\n        )\n\n    def get_transaction(\n        self, filename: str, index: int, recipient_account: str\n    ) -> data.Transaction:\n        return data.Transaction(\n            meta=data.new_metadata(filename, index),\n            date=parse(self.data["Posted Date"]),\n            flag=flags.FLAG_OKAY,\n            payee=self.data["Payee"],\n            narration="",\n            tags=set(),\n            links=set(),\n            postings=[\n                data.Posting(\n                    "Liabilities:BofaCreditCard",\n                    amount.Amount(D(self.data["Amount"]), "USD"),\n                    None,\n                    None,\n                    None,\n                    None,\n                ),\n                data.Posting(recipient_account, None, None, None, None, None),\n            ],\n        )\n\n\nclass BofaCCImporter(PromptImporter):\n    def __init__(self, db_file):\n        super().__init__(db_file)\n\n    def name(self):\n        return "BofaCCImporter"\n\n    def get_events(self, f) -> list[Event]:\n        with open(f.name) as infile:\n            return [BofaCCEvent(row) for row in csv.DictReader(infile)]\n```\n',
    'author': 'Harry Eldridge',
    'author_email': 'eldridgemharry@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/heldridge/prompt_importer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
